r'''
Base kernel classes for Gaussian Process Regression fitting of 1D data with errorbars. Built in Python 3.x, adapted to be Python 2.x compatible.
  06/12/2024: No longer compatible with Python 2.x.
'''

# Required imports
import warnings
import numpy as np

from .definitions import number_types, array_types, default_dtype


class _Kernel():
    r'''
    Base class to be inherited by **ALL** kernel implementations in order for type checks to succeed.

    Ideology:

    - :code:`self._fname` is a string, designed to provide an easy way to check the kernel instance type.
    - :code:`self._function` contains the covariance function, k, along with **at least** dk/dx1, dk/dx2, and d^2k/dx1dx2.
    - :code:`self._hyperparameters` contains free variables that are designed to vary in logarithmic-space.
    - :code:`self._constants` contains free variables that should not be changed during parameter searches, or true constants.
    - :code:`self._bounds` contains the bounds of the free variables to be used in randomized kernel restart algorithms.

    .. tip::

        Get/set functions implemented as class properties now (v >= 2.0.0)

    :kwarg name: str. Codename of :code:`_Kernel` class implementation.

    :kwarg func: callable. Covariance function of :code:`_Kernel` class implementation.

    :kwarg hderf: bool. Indicates availability of analytical hyperparameter derivatives within :code:`func` for optimization algorithms.

    :kwarg hyps: array. Hyperparameters to be stored in the :code:`_Kernel` instance, ordered according to the specific :code:`_Kernel` class implementation.

    :kwarg csts: array. Constants to be stored in the :code:`_Kernel` instance, ordered according to the specific :code:`_Kernel` class implementation.

    :kwarg htags: array. Names of hyperparameters to be stored as indices in the :code:`_Kernel` class implementation. (optional)

    :kwarg ctags: array. Names of constants to be stored as indices in the :code:`_Kernel` class implementation. (optional)
    '''

    def __init__(
        self,
        name=None,
        func=None,
        hderf=False,
        hyps=None,
        csts=None,
        htags=None,
        ctags=None,
        dtype=None
    ):
        r'''
        Initializes the :code:`_Kernel` instance.

        .. note::

            Nothing is done with the :code:`htags` and :code:`ctags` arguments currently. (v >= 1.0.1)

        :kwarg name: str. Codename of :code:`_Kernel` class implementation.

        :kwarg func: callable. Covariance function of :code:`_Kernel` class implementation.

        :kwarg hderf: bool. Indicates availability of analytical hyperparameter derivatives within :code:`func` for optimization algorithms.

        :kwarg hyps: array. Hyperparameters to be stored in the :code:`_Kernel` instance, ordered according to the specific :code:`_Kernel` class implementation.

        :kwarg csts: array. Constants to be stored in the :code:`_Kernel` instance, ordered according to the specific :code:`_Kernel` class implementation.

        :kwarg htags: array. Names of hyperparameters to be stored as indices in the :code:`_Kernel` class implementation. (optional)

        :kwarg ctags: array. Names of constants to be stored as indices in the :code:`_Kernel` class implementation. (optional)

        :returns: none.
        '''

        self._dtype = dtype if dtype is not None else default_dtype
        self._fname = name
        self._function = func if callable(func) else None
        self._hyperparameters = np.array(hyps, dtype=self._dtype).flatten() if isinstance(hyps, array_types) else None
        self._constants = np.array(csts, dtype=self._dtype).flatten() if isinstance(csts, array_types) else None
        self._hyp_lbounds = None
        self._hyp_ubounds = None
        self._hderflag = hderf
        self._force_bounds = False


    def __call__(self, x1, x2, der=0, hder=None):
        r'''
        Default class call function, evaluates the stored covariance function at the input values.

        :arg x1: array. Meshgrid of x_1-values at which to evaulate the covariance function.

        :arg x2: array. Meshgrid of x_2-values at which to evaulate the covariance function.

        :kwarg der: int. Order of x derivative with which to evaluate the covariance function, requires explicit implementation. (optional)

        :kwarg hder: int. Order of hyperparameter derivative with which to evaluate the covariance function, requires explicit implementation. (optional)

        :returns: array. Covariance function evaluations at input value pairs using the given derivative settings. Has the same dimensions as :code:`x1` and :code:`x2`.
        '''

        k_out = None
        if callable(self._function):
            xt1 = None
            xt2 = None
            dert = 0
            hdert = None
            if isinstance(x1, number_types):
                xt1 = np.array(np.atleast_1d(x1), dtype=self._dtype)
            elif isinstance(x1, array_types):
                xt1 = np.array(np.atleast_1d(x1), dtype=self._dtype)
            if isinstance(x2, number_types):
                xt2 = np.array(np.atleast_1d(x2), dtype=self._dtype)
            elif isinstance(x2, array_types):
                xt2 = np.array(np.atleast_1d(x2), dtype=self._dtype)
            if isinstance(der, number_types):
                dert = int(der)
            if isinstance(hder, number_types):
                hdert = int(hder)
            if isinstance(xt1, np.ndarray) and isinstance(xt2, np.ndarray):
                k_out = self._function(xt1, xt2, dert, hdert)
            else:
                raise TypeError(f'Arguments x1 and x2 must be at least 1D-array-like object.')
        else:
            raise NotImplementedError(f'Covariance function of {self.name} Kernel object not yet defined.')
        return k_out


    def __eq__(self,other):
        r'''
	Custom equality operator. Compares name, hyperparameters and constants.

        :arg other: obj. Any other :code:`_Kernel` class instance.

        :returns: bool. Indicates whether the two objects are equal to each other.
        '''

        status = False
        if isinstance(other, _Kernel):
            if self.name == other.name:
                shyp = np.all(np.isclose(self.hyperparameters, other.hyperparameters))
                scst = np.all(np.isclose(self.constants, other.constants))
                status = self.hyperparameters.size == other.hyperparameters.size and self.constants.size == other.constants.size and shyp and scst
        return status


    def __ne__(self, other):
        r'''
	Custom inequality operator. Compares name, hyperparameters and constants.

        :arg other: obj. Any other :code:`_Kernel` class instance.

        :returns: bool. Indicates whether the two objects are not equal to each other.
        '''

        return not self.__eq__(other)


    def enforce_bounds(self, value=True):
        r'''
        Sets a flag to enforce the given hyperparameter bounds.

        :kwarg value: bool. Boolean value to set the flag.

        :returns: none.
        '''

        self._force_bounds = True if value else False


    @property
    def name(self):
        r'''
        Returns the codename of the :code:`_Kernel` instance.

        :returns: str. Codename of the :code:`_Kernel` instance.
        '''

        return f'{self._fname}'


    @property
    def hyperparameters(self):
        r'''
        Return the hyperparameters stored in the :code:`_Kernel` instance.

        :returns: array. Hyperparameter list, ordered according to the specific :code:`_Kernel` class implementation.
        '''

        val = np.array([])
        if isinstance(self._hyperparameters, np.ndarray):
            val = self._hyperparameters.copy()
        return val


    @property
    def constants(self):
        r'''
        Return the constants stored in the :code:`_Kernel` instance.

        :returns: array. Constant list, ordered according to the specific :code:`_Kernel` class implementation.
        '''

        val = np.array([])
        if isinstance(self._constants, np.ndarray):
            val = self._constants.copy()
        return val


    @property
    def bounds(self):
        r'''
        Return the hyperparameter search bounds stored in the :code:`_Kernel` instance.

        :returns: array. Hyperparameter lower/upper bounds list, ordered according to the specific :code:`_Kernel` class implementation.
        '''

        val = None
        if isinstance(self._hyp_lbounds, np.ndarray) and isinstance(self._hyp_ubounds, np.ndarray) and self._hyp_lbounds.shape == self._hyp_ubounds.shape:
            val = np.vstack((self._hyp_lbounds.flatten(), self._hyp_ubounds.flatten()))
        return val


    def is_hderiv_implemented(self):
        r'''
        Checks if the explicit hyperparameter derivative is implemented in the :code:`_Kernel` class implementation.

        :returns: bool. True if explicit hyperparameter derivative is implemented.
        '''

        return self._hderflag


    @hyperparameters.setter
    def hyperparameters(self, theta):
        r'''
        Set the hyperparameters stored in the :code:`_Kernel` instance.

        :arg theta: array. Hyperparameter list to be stored, ordered according to the specific :code:`_Kernel` implementation.

        :returns: none.
        '''

        userhyps = None
        if isinstance(theta, array_types):
            userhyps = np.array(theta, dtype=self._dtype).flatten()
        else:
            raise TypeError(f'{self.name} Kernel hyperparameters must be given as an array-like object.')
        nhyps = self._hyperparameters.size
        if nhyps > 0:
            if userhyps.size >= nhyps:
                if self._force_bounds and isinstance(self._hyp_lbounds, np.ndarray) and self._hyp_lbounds.size == nhyps:
                    htemp = userhyps[:nhyps]
                    lcheck = (htemp < self._hyp_lbounds)
                    htemp[lcheck] = self._hyp_lbounds[lcheck]
                    userhyps[:nhyps] = htemp
                if self._force_bounds and isinstance(self._hyp_ubounds, np.ndarray) and self._hyp_ubounds.size == nhyps:
                    htemp = userhyps[:nhyps]
                    ucheck = (htemp > self._hyp_ubounds)
                    htemp[ucheck] = self._hyp_ubounds[ucheck]
                    userhyps[:nhyps] = htemp
                self._hyperparameters = np.array(userhyps[:nhyps], dtype=self._dtype)
            else:
                raise ValueError(f'{self.name} Kernel hyperparameters must contain at least {nhyps} elements.')
        else:
            warnings.warn(f'{self.name} Kernel instance has no hyperparameters.', stacklevel=2)


    @constants.setter
    def constants(self,consts):
        r'''
        Set the constants stored in the :code:`_Kernel` instance.

        :arg consts: array. Constant list to be stored, ordered according to the specific :code:`_Kernel` class implementation.

        :returns: none.
        '''

        usercsts = None
        if isinstance(consts, array_types):
            usercsts = np.array(consts, dtype=self._dtype).flatten()
        else:
            raise TypeError(f'{self.name} Kernel constants must be given as an array-like object.')
        ncsts = self._constants.size
        if ncsts > 0:
            if usercsts.size >= ncsts:
                self._constants = np.array(usercsts[:ncsts], dtype=self._dtype)
            else:
                raise ValueError(f'{self.name} Kernel constants must contain at least {ncsts} elements.')
        else:
            warnings.warn(f'{self.name} Kernel instance has no constants.', stacklevel=2)


    @bounds.setter
    def bounds(self,bounds):
        r'''
        Set the hyperparameter bounds stored in the :code:`_Kernel` instance.

        :arg bounds: 2D array. Hyperparameter lower/upper bound list to be stored, ordered according to the specific :code:`_Kernel` class implementation.

        :returns: none.
        '''

        userbnds = None
        if isinstance(bounds, array_types):
            userbnds = np.atleast_2d(np.array(bounds, dtype=self._dtype))
        else:
            raise TypeError(f'{self.name} Kernel bounds must be given as a 2D-array-like object with exactly 2 rows.')
        if userbnds.shape[0] != 2:
            raise TypeError(f'{self.name} Kernel bounds must be given as a 2D-array-like object with exactly 2 rows.')
        nhyps = self._hyperparameters.size
        if nhyps > 0:
            if userbnds.shape[1] >= nhyps:
                self._hyp_lbounds = np.array(userbnds[0, :nhyps], dtype=self._dtype)
                self._hyp_ubounds = np.array(userbnds[1, :nhyps], dtype=self._dtype)
                if self._force_bounds:
                    self.hyperparameters = self._hyperparameters.copy()
            else:
                raise ValueError(f'{self.name} Kernel bounds must be a 2D-array-like object with exactly 2 rows and at least {nhyps} elements per row.')
        else:
            warnings.warn(f'{self.name} Kernel instance has no hyperparameters to set bounds for.', stacklevel=2)



class _OperatorKernel(_Kernel):
    r'''
    Base operator class to be inherited by **ALL** operator kernel implementations for custom get/set functions.
    Type checking done with :code:`isinstance(<object>,<this_module>._OperatorKernel)` if needed.

    Ideology:

    - :code:`self._kernel_list` is a Python list of :code:`_Kernel` instances on which the specified operation will be performed

    Get/set functions adjusted to call get/set functions of each constituent kernel instead of using its own
    attributes, which are mostly left as :code:`None`.

    .. warning::

        Get/set functions not yet programmed as actual attributes! May required some code restructuring to incorporate
        this though. (v >= 1.0.1)

    :kwarg name: str. Codename of :code:`_OperatorKernel` class implementation.

    :kwarg func: callable. Covariance function of :code:`_OperatorKernel` class implementation, ideally an operation on provided :code:`_Kernel` instances.

    :kwarg hderf: bool. Indicates availability of analytical hyperparameter derivatives within :code:`func` for optimization algorithms.

    :kwarg klist: array. List of :code:`_Kernel` instances to be operated on by the :code:`_OperatorKernel` instance, input order determines the order of parameter lists.
    '''

    def __init__(
        self,
        name='Op',
        func=None,
        hderf=False,
        klist=None,
        dtype=None
    ):
        r'''
        Initializes the :code:`_OperatorKernel` instance.

        :kwarg name: str. Codename of :code:`_OperatorKernel` class implementation.

        :kwarg func: callable. Covariance function of :code:`_OperatorKernel` class implementation, ideally an operation on provided :code:`_Kernel` instances.

        :kwarg hderf: bool. Indicates availability of analytical hyperparameter derivatives within :code:`func` for optimization algorithms.

        :kwarg klist: array. List of :code:`_Kernel` instances to be operated on by the :code:`_OperatorKernel` instance, input order determines the order of parameter lists.

        :returns: none.
        '''

        super().__init__(name=name, func=func, hderf=hderf, dtype=dtype)
        self._kernel_list = klist if klist is not None else []


    @property
    def name(self):
        r'''
        Returns the codename of the :code:`_OperatorKernel` instance.

        :returns: str. Codename of the :code:`_OperatorKernel` instance.
        '''

        kname = '-'.join([f'{kk.name}' for kk in self._kernel_list])
        return f'{self._fname}({kname})'


    @property
    def basename(self):
        r'''
        Returns the base codename of the :code:`_OperatorKernel` instance.

        :returns: str. Base codename of the :code:`_OperatorKernel` instance.
        '''

        return f'{self._fname}'


    @property
    def hyperparameters(self):
        r'''
        Return the hyperparameters of all the :code:`_Kernel` instances stored within the :code:`_OperatorKernel` instance.

        :returns: array. Hyperparameter list, ordered according to the specific :code:`_OperatorKernel` class implementation and the current :code:`self._kernel_list` instance.
        '''

        val = np.array([])
        for kk in self._kernel_list:
            val = np.hstack((val, kk.hyperparameters))
        return val


    @property
    def constants(self):
        r'''
        Return the constants of all the :code:`_Kernel` instances stored within the :code:`_OperatorKernel` instance.

        :returns: array. Constant list, ordered according to the specific :code:`_OperatorKernel` class implementation and the current :code:`self._kernel_list` instance.
        '''

        val = np.array([])
        for kk in self._kernel_list:
            val = np.hstack((val, kk.constants))
        return val


    @property
    def bounds(self):
        r'''
        Return the hyperparameter bounds of all the :code:`_Kernel` instances stored within the :code:`_OperatorKernel` instance.

        :returns: array. Hyperparameter lower/upper bounds list, ordered according to the specific :code:`_OperatorKernel` class implementation and the current :code:`self._kernel_list` instance.
        '''

        val = None
        for kk in self._kernel_list:
            kval = kk.bounds
            if kval is not None:
                val = np.hstack((val, kval)) if val is not None else kval.copy()
        return val


    @hyperparameters.setter
    def hyperparameters(self, theta):
        r'''
        Set the hyperparameters stored in all the :code:`_Kernel` instances within the :code:`_OperatorKernel` instance.

        :arg theta: array. Hyperparameter list to be stored, ordered according to the specific :code:`_OperatorKernel` class implementation and the current :code:`self._kernel_list` instance.

        :returns: none.
        '''

        userhyps = None
        if isinstance(theta, array_types):
            userhyps = np.array(theta).flatten()
        else:
            raise TypeError(f'{self.name} OperatorKernel hyperparameters must be given as an array-like object.')
        nhyps = self.hyperparameters.size
        if nhyps > 0:
            if userhyps.size >= nhyps:
                ndone = 0
                for kk in self._kernel_list:
                    nhere = ndone + kk.hyperparameters.size
                    if nhere != ndone:
                        if nhere == nhyps:
                            kk.hyperparameters = theta[ndone:]
                        else:
                            kk.hyperparameters = theta[ndone:nhere]
                        ndone = nhere
            else:
                raise ValueError(f'{self.name} OperatorKernal hyperparameters must contain at least {nhyps} elements.')
        else:
            warnings.warn(f'{self.name} OperatorKernel instance has no hyperparameters.', stacklevel=2)


    @constants.setter
    def constants(self, consts):
        r'''
        Set the constants stored in all the :code:`_Kernel` instances within the :code:`_OperatorKernel` instance.

        :arg consts: array. Constant list to be stored, ordered according to the specific :code:`_OperatorKernel` class implementation and the current :code:`self._kernel_list` instance.

        :returns: none.
        '''

        usercsts = None
        if isinstance(consts, array_types):
            usercsts = np.array(consts).flatten()
        else:
            raise TypeError(f'{self.name} OperatorKernel constants must be given as an array-like object.')
        ncsts = self.constants.size
        if ncsts > 0:
            if usercsts.size >= ncsts:
                ndone = 0
                for kk in self._kernel_list:
                    nhere = ndone + kk.constants.size
                    if nhere != ndone:
                        if nhere == ncsts:
                            kk.constants = consts[ndone:]
                        else:
                            kk.constants = consts[ndone:nhere]
                        ndone = nhere
            else:
                raise ValueError(f'{self.name} OperatorKernel constants must contain at least {ncsts} elements.')
        else:
            warnings.warn(f'{self.name} OperatorKernel instance has no constants.', stacklevel=2)


    @bounds.setter
    def bounds(self, bounds):
        r'''
        Set the hyperparameter bounds stored in all the :code:`_Kernel` instances within the :code:`_OperatorKernel` instance.

        :arg bounds: array. Hyperparameter lower/upper bound list to be stored, ordered according to the specific :code:`_OperatorKernel` class implementation and the current :code:`self._kernel_list` instance.

        :returns: none.
        '''

        userbnds = None
        if isinstance(bounds, array_types):
            userbnds = np.atleast_2d(bounds)
        else:
            raise TypeError(f'{self.name} OperatorKernel bounds must be given as a 2d-array-like object with exactly 2 rows.')
        if userbnds.shape[0] != 2:
            raise TypeError(f'{self.name} OperatorKernel bounds must be given as a 2d-array-like object with exactly 2 rows.')
        nhyps = self.hyperparameters.size
        if nhyps > 0:
            if userbnds.shape[1] >= nhyps:
                ndone = 0
                for kk in self._kernel_list:
                    nhere = ndone + kk.hyperparameters.size
                    if nhere != ndone:
                        if nhere == nhyps:
                            kk.bounds = userbnds[:, ndone:]
                        else:
                            kk.bounds = userbnds[:, ndone:nhere]
                        ndone = nhere
            else:
                raise ValueError(f'{self.name} OperatorKernel bounds must be a 2D-array-like object with exactly 2 rows and contain at least {nhyps} elements per row.')
        else:
            warnings.warn(f'{self.name} OperatorKernel instance has no hyperparameters to set bounds for.', stacklevel=2)



class _WarpingFunction():
    r'''
    Base class to be inherited by **ALL** warping function implementations in order for type checks to succeed.
    Type checking done with :code:`isinstance(<object>,<this_module>._WarpingFunction)`.

    Ideology:

    - :code:`self._fname` is a string, designed to provide an easy way to check the warping function instance type.
    - :code:`self._function` contains the warping function, l, along with *at least* dl/dz and d^2l/dz^2.
    - :code:`self._hyperparameters` contains free variables that are designed to vary in logarithmic-space.
    - :code:`self._constants` contains free variables that should not be changed during hyperparameter optimization, or true constants.
    - :code:`self._bounds` contains the bounds of the free variables to be used in randomized kernel restarts.

    Get/set functions already given, but as always in Python, all functions can be overridden by specific implementation.
    This is strongly **NOT** recommended unless you are familiar with how these structures work and their interdependencies.

    .. warning::

        Get/set functions not yet programmed as actual attributes! May required some code restructuring to incorporate
        this though. (v >= 1.0.1)

    .. note::

        The usage of the variable z in the documentation is simply to emphasize the generality of the object. In actuality,
        it is the same as x within the :code:`_Kernel` base class.

    :kwarg name: str. Codename of :code:`_WarpingFunction` class implementation.

    :kwarg func: callable. Warping function of :code:`_WarpingFunction` class implementation.

    :kwarg hderf: bool. Indicates availability of analytical hyperparameter derivatives within :code:`func` for optimization algorithms.

    :kwarg hyps: array. Hyperparameters to be stored in the :code:`_WarpingFunction` instance, ordered according to the specific :code:`_WarpingFunction` class implementation.

    :kwarg csts: array. Constants to be stored in the :code:`_WarpingFunction` instance, ordered according to the specific :code:`_WarpingFunction` class implementation.

    :kwarg htags: array. Names of hyperparameters to be stored as indices in the :code:`_WarpingFunction` class implementation. (optional)

    :kwarg ctags: array. Names of constants to be stored as indices in the :code:`_WarpingFunction` class implementation. (optional)
    '''

    def __init__(
        self,
        name='Warp',
        func=None,
        hderf=False,
        hyps=None,
        csts=None,
        htags=None,
        ctags=None,
        dtype=None
    ):
        r'''
        Initializes the :code:`_WarpingFunction` instance.

        .. note::

            Nothing is done with the :code:`htags` and :code:`ctags` arguments currently. (v >= 1.0.1)

        :kwarg name: str. Codename of :code:`_WarpingFunction` class implementation.

        :kwarg func: callable. Warping function of :code:`_WarpingFunction` class implementation.

        :kwarg hderf: bool. Indicates availability of analytical hyperparameter derivatives within :code:`func` for optimization algorithms.

        :kwarg hyps: array. Hyperparameters to be stored in the :code:`_WarpingFunction` instance, ordered according to the specific :code:`_WarpingFunction` class implementation.

        :kwarg csts: array. Constants to be stored in the :code:`_WarpingFunction` instance, ordered according to the specific :code:`_WarpingFunction` class implementation.

        :kwarg htags: array. Names of hyperparameters to be stored as indices in the :code:`_WarpingFunction` class implementation. (optional)

        :kwarg ctags: array. Names of constants to be stored as indices in the :code:`_WarpingFunction` class implementation. (optional)

        :returns: none.
        '''

        self._dtype = dtype if dtype is not None else default_dtype
        self._fname = name
        self._function = func if func is not None else None
        self._hyperparameters = np.array(hyps, dtype=self._dtype) if isinstance(hyps, array_types) else None
        self._constants = np.array(csts, dtype=self._dtype) if isinstance(csts, array_types) else None
        self._hyp_lbounds = None
        self._hyp_ubounds = None
        self._hderflag = hderf
        self._force_bounds = False


    def __call__(self, zz, der=0, hder=None):
        r'''
        Default class call function, evaluates the stored warping function at the input values.

        :arg zz: array. Vector of z-values at which to evaulate the warping function, can be 1D or 2D depending on application.

        :kwarg der: int. Order of z derivative with which to evaluate the warping function, requires explicit implementation. (optional)

        :kwarg hder: int. Order of hyperparameter derivative with which to evaluate the warping function, requires explicit implementation. (optional)

        :returns: array. Warping function evaluations at input values using the given derivative settings. Has the same dimensions as :code:`zz`.
        '''

        k_out = None
        if self._function is not None:
            k_out = self._function(zz, der, hder)
        else:
            raise NotImplementedError('Warping function not yet defined.')
        return k_out


    def __eq__(self, other):
        r'''
	Custom equality operator. Compares name, hyperparameters and constants.

        :arg other: obj. Any other :code:`_WarpingFunction` class instance.

        :returns: bool. Indicates whether the two objects are equal to each other.
        '''

        status = False
        if isinstance(other, _WarpingFunction):
            if self.name == other.name:
                shyp = np.all(np.isclose(self.hyperparameters, other.hyperparameters))
                scst = np.all(np.isclose(self.constants, other.constants))
                status = self.hyperparameters.size == other.hyperparameters.size and self.constants.size == other.constants.size and shyp and scst
        return status


    def __ne__(self, other):
        r'''
	Custom inequality operator. Compares name, hyperparameters and constants.

        :arg other: obj. Any other :code:`_WarpingFunction` class instance.

        :returns: bool. Indicates whether the two objects are not equal to each other.
        '''

        return not self.__eq__(other)


    def enforce_bounds(self, value=True):
        r'''
        Sets a flag to enforce the given hyperparameter bounds.

        :kwarg value: bool. Boolean value to set the flag.

        :returns: none.
        '''

        self._force_bounds = True if value else False


    @property
    def name(self):
        r'''
        Returns the codename of the :code:`_WarpingFunction` instance.

        :returns: str. Codename of the :code:`_WarpingFunction` instance.
        '''

        return f'{self._fname}'


    @property
    def hyperparameters(self):
        r'''
        Return the hyperparameters stored in the :code:`_WarpingFunction` instance.

        :returns: array. Hyperparameter list, ordered according to the specific :code:`_WarpingFunction` class implementation.
        '''

        val = np.array([])
        if self._hyperparameters is not None:
            val = self._hyperparameters.copy()
        return val


    @property
    def constants(self):
        r'''
        Return the constants stored in the :code:`_WarpingFunction` instance.

        :returns: array. Constant list, ordered according to the specific :code:`_WarpingFunction` class implementation.
        '''

        val = np.array([])
        if self._constants is not None:
            val = self._constants.copy()
        return val


    @property
    def bounds(self):
        r'''
        Return the hyperparameter search bounds stored in the :code:`_WarpingFunction` instance.

        :returns: array. Hyperparameter bounds list, ordered according to the specific :code:`_WarpingFunction` class implementation.
        '''

        val = None
        if isinstance(self._hyp_lbounds, np.ndarray) and isinstance(self._hyp_ubounds, np.ndarray) and self._hyp_lbounds.shape == self._hyp_ubounds.shape:
            val = np.vstack((self._hyp_lbounds.flatten(), self._hyp_ubounds.flatten()))
        return val


    def is_hderiv_implemented(self):
        r'''
        Checks if the explicit hyperparameter derivative is implemented in this :code:`_WarpingFunction` class implementation.

        :returns: bool. True if explicit hyperparameter derivative is implemented.
        '''

        return self._hderflag


    @hyperparameters.setter
    def hyperparameters(self, theta):
        r'''
        Set the hyperparameters stored in the :code:`_WarpingFunction` instance.

        :arg theta: array. Hyperparameter list to be stored, ordered according to the specific :code:`_WarpingFunction` class implementation.

        :returns: none.
        '''

        userhyps = None
        if isinstance(theta, array_types):
            userhyps = np.array(theta, dtype=self._dtype).flatten()
        else:
            raise TypeError(f'{self.name} WarpingFunction hyperparameters must be given as an array-like object.')
        nhyps = self.hyperparameters.size
        if nhyps > 0:
            if userhyps.size >= nhyps:
                if self._force_bounds and isinstance(self._hyp_lbounds, np.ndarray) and self._hyp_lbounds.size == nhyps:
                    htemp = userhyps[:nhyps]
                    lcheck = (htemp < self._hyp_lbounds)
                    htemp[lcheck] = self._hyp_lbounds[lcheck]
                    userhyps[:nhyps] = htemp
                if self._force_bounds and isinstance(self._hyp_ubounds, np.ndarray) and self._hyp_ubounds.size == nhyps:
                    htemp = userhyps[:nhyps]
                    ucheck = (htemp > self._hyp_ubounds)
                    htemp[ucheck] = self._hyp_ubounds[ucheck]
                    userhyps[:nhyps] = htemp
                self._hyperparameters = np.array(userhyps[:nhyps], dtype=self._dtype)
            else:
                raise ValueError(f'{self.name} WarpingFunction hyperparameters must contain at least {nhyps} elements.')
        else:
            warnings.warn(f'{self.name} WarpingFunction instance has no hyperparameters.', stacklevel=2)


    @constants.setter
    def constants(self, consts):
        r'''
        Set the constants stored in the :code:`_WarpingFunction` object.

        :arg consts: array. Constant list to be stored, ordered according to the specific :code:`_WarpingFunction` class implementation.

        :returns: none.
        '''

        usercsts = None
        if isinstance(consts, array_types):
            usercsts = np.array(consts, dtype=self._dtype).flatten()
        else:
            raise TypeError(f'{self.name} WarpingFunction constants must be given as an array-like object.')
        ncsts = self.constants.size
        if ncsts > 0:
            if usercsts.size >= ncsts:
                self._constants = np.array(usercsts[:ncsts], dtype=self._dtype)
            else:
                raise ValueError(f'{self.name} WarpingFunction constants must contain at least {ncsts} elements.')
        else:
            warnings.warn(f'{self.name} WarpingFunction instance has no constants.', stacklevel=2)


    @bounds.setter
    def bounds(self,bounds):
        r'''
        Set the hyperparameter bounds stored in the :code:`_WarpingFunction` instance.

        :arg bounds: array. Hyperparameter lower/upper bound list to be stored, ordered according to the specific :code:`_WarpingFunction` class implementation.

        :returns: none.
        '''

        userbnds = None
        if isinstance(bounds, array_types):
            userbnds = np.atleast_2d(np.array(bounds, dtype=self._dtype))
        else:
            raise TypeError(f'{self.name} WarpingFunction bounds must be given as a 2D-array-like object with exactly 2 rows.')
        if userbnds.shape[0] != 2:
            raise TypeError(f'{self.name} WarpingFunction bounds must be given as a 2D-array-like object with exactly 2 rows.')
        nhyps = self.hyperparameters.size
        if nhyps > 0:
            if userbnds.shape[1] >= nhyps:
                self._hyp_lbounds = np.array(userbnds[0, :nhyps], dtype=self._dtype)
                self._hyp_ubounds = np.array(userbnds[1, :nhyps], dtype=self._dtype)
                if self._force_bounds:
                    self.hyperparameters = self._hyperparameters.copy()
            else:
                raise ValueError(f'{self.name} WarpingFunction bounds must be a 2D-array-like object with exactly 2 rows and contain at least {nhyps} elements per row.')
        else:
            warnings.warn(f'{self.name} WarpingFunction instance has no hyperparameters to set bounds for.', stacklevel=2)




r'''
A simplified class for easy access to new users not needing fine control
'''

# Required imports
import numpy as np

from .definitions import number_types, array_types
from .kernels import RQ_Kernel
from .routines import GaussianProcess

__all__ = [
    'SimplifiedGaussianProcess',  # Simplified fitting routine wrapper
]


class SimplifiedGaussianProcess(GaussianProcess):
    r'''
    A simplified version of the main :code:`GaussianProcess`
    class with pre-defined settings, only requiring the bare necessities
    to use for fitting. Although this class allows an entry point akin
    to typical :mod:`scipy` classes, it is provided primarily to be
    used as a template for implementations meant to simplify the :code:`mkgp`
    experience for the average user.

    .. note::

        Optimization of hyperparameters is only performed *once* using
        settings at the time of the *first* call! All subsequent calls
        use the results of the first optimization, regardless of its
        quality or convergence status.

    :arg kernel: object. The covariance function, as a :code:`_Kernel` instance, to be used in the fit procedure.

    :arg xdata: array. Vector of x-values corresponding to data to be fitted.

    :arg ydata: array. Vector of y-values corresponding to data to be fitted. Must be the same shape as :code:`xdata`.

    :arg yerr: array. Vector of y-errors corresponding to data to be fitted, assumed to be given as 1 sigma. Must be the same shape as :code:`xdata`.

    :kwarg xerr: array. Vector of x-errors corresponding to data to be fitted, assumed to be given as 1 sigma. Must be the same shape as :code:`xdata`. (optional)

    :kwarg kernel_bounds: array. 2D array with rows being hyperparameters and columns being :code:`[lower,upper]` bounds. (optional)

    :kwarg reg_par: float. Parameter adjusting penalty on kernel complexity. (optional)

    :kwarg epsilon: float. Convergence criterion on change in log-marginal-likelihood. (optional)

    :kwarg num_restarts: int. Number of kernel restarts. (optional)

    :kwarg hyp_opt_gain: float. Gain value on the hyperparameter optimizer, expert use only. (optional)
    '''

    def __init__(
        self,
        kernel,
        xdata,
        ydata,
        yerr,
        xerr=None,
        kernel_bounds=None,
        reg_par=1.0,
        epsilon=1.0e-2,
        num_restarts=0,
        hyp_opt_gain=1.0e-2,
        include_noise=True
    ):
        r'''
        Defines customized :code:`GaussianProcess` instance with
        a pre-defined common settings for both data fit and error fit. Input
        parameters reduced only to essentials and most crucial knobs for
        fine-tuning.

        :arg kernel: object. The covariance function, as a :code:`_Kernel` instance, to be used in the fit procedure.

        :arg xdata: array. Vector of x-values corresponding to data to be fitted.

        :arg ydata: array. Vector of y-values corresponding to data to be fitted. Must be the same shape as :code:`xdata`.

        :arg yerr: array. Vector of y-errors corresponding to data to be fitted, assumed to be given as 1 sigma. Must be the same shape as :code:`xdata`.

        :kwarg xerr: array. Optional vector of x-errors corresponding to data to be fitted, assumed to be given as 1 sigma. Must be the same shape as :code:`xdata`. (optional)

        :kwarg kernel_bounds: array. 2D array with rows being hyperparameters and columns being :code:`[lower,upper]` bounds. (optional)

        :arg reg_par: float. Parameter adjusting penalty on kernel complexity. (optional)

        :kwarg epsilon: float. Convergence criterion on change in log-marginal-likelihood. (optional)

        :kwarg num_restarts: int. Number of kernel restarts. (optional)

        :kwarg hyp_opt_gain: float. Gain value on the hyperparameter optimizer, expert use only. (optional)

        :kwarg include_noise: bool. Specifies inclusion of Gaussian noise term in returned variance. Only operates on diagonal elements. (optional)

        :returns: none.
        '''
        super().__init__()
        self._nrestarts = num_restarts

        self.set_raw_data(xdata=xdata, ydata=ydata, yerr=yerr, xerr=xerr)

        eps = 'none' if not isinstance(epsilon, number_types) else epsilon
        sg = hyp_opt_gain if isinstance(hyp_opt_gain, number_types) else 1.0e-1
        self.set_kernel(kernel=kernel, kbounds=kernel_bounds, regpar=reg_par)
        self.set_search_parameters(epsilon=epsilon, method='adam', spars=[sg, 0.4, 0.8])

        self._perform_heterogp = False if isinstance(yerr, number_types) or self._ye is None else True
        self._perform_nigp = False if self._xe is None else True
        self._include_noise = True if include_noise else False

        if self._perform_heterogp:
            error_length = 5.0 * (np.nanmax(self._xx) - np.nanmin(self._xx)) / float(self._xx.size) if self._xx is not None else 5.0e-1
            error_kernel = RQ_Kernel(5.0e-1, error_length, 3.0e1)
            error_kernel_hyppar_bounds = np.atleast_2d([[1.0e-1, error_length * 0.25, 1.0e1], [1.0e0, error_length * 4.0, 5.0e1]])
            self.set_error_kernel(kernel=error_kernel, kbounds=error_kernel_hyppar_bounds, regpar=5.0, nrestarts=0)
            self.set_error_search_parameters(epsilon=1.0e-1, method='adam', spars=[1.0e-2, 0.4, 0.8])
        elif isinstance(yerr, number_types):
            ye = np.full(self._yy.shape, yerr)
            self.set_raw_data(yerr=ye)


    def __call__(self, xnew):
        r'''
        Defines a simplified fitting execution, *only* performs
        optimization on the *first* call. Subsequent calls
        merely evaluate the optimized fit at the input x-values.

        :arg xnew: array. Vector of x-values corresponding to points at which the GPR results should be evaulated.

        :returns: tuple.
                  Mean of GPR predictive distribution, ie. the fit ;
                  Standard deviation of mean, given as 1 sigma ;
                  Mean derivative of GPR predictive disstribution, ie. the derivative of the fit ;
                  Standard deviation of mean derivative, given as 1 sigma.
        '''
        nrestarts = self._nrestarts
        if self._xF is not None:
            nrestarts = None
            self.set_search_parameters(epsilon='none')
            self.set_error_search_parameters(epsilon='none')
        self.GPRFit(xnew, hsgp_flag=self._perform_heterogp, nigp_flag=self._perform_nigp, nrestarts=nrestarts)
        return self.get_gp_results(noise_flag=self._include_noise)


    def sample(self, xnew, derivative=False):
        r'''
        Provides a more intuitive function for sampling the
        predictive distribution. Only provides one sample
        per call, unlike the more complex function in the
        main class.

        :arg xnew: array. Vector of x-values corresponding to points where GPR results should be evaulated at.

        :kwarg derivative: bool. Flag to indicate sampling of fit derivative instead of the fit. (optional)

        :returns: array. Vector of y-values corresponding to a random sample of the GPR predictive distribution.
        '''
        self.__call__(xnew)
        remove_noise = not self._include_noise
        output = self.sample_GP(1, actual_noise=False, without_noise=remove_noise) if not derivative else self.sample_GP_derivative(1, actual_noise=False, without_noise=remove_noise)
        return output




r'''
Some helpful functions and classes for reproducability and user-friendliness
'''

# Required imports
import re
import numpy as np

from .definitions import number_types, array_types, default_dtype
from .kernels import (
    _Kernel,
    Constant_Kernel,
    Noise_Kernel,
    Linear_Kernel,
    Poly_Order_Kernel,
    SE_Kernel,
    RQ_Kernel,
    Matern_HI_Kernel,
    NN_Kernel,
    Gibbs_Kernel,
    Sum_Kernel,
    Product_Kernel,
    Symmetric_Kernel,
    ND_Sum_Kernel,
    ND_Product_Kernel,
    Constant_WarpingFunction,
    Linear_WarpingFunction,
    IG_WarpingFunction,
)

__all__ = [
    'KernelConstructor', 'KernelReconstructor',  # Kernel construction functions
]


def KernelConstructor(name, dtype=None):
    r'''
    Function to construct a kernel solely based on the kernel codename.

    .. note::

        All :code:`_OperatorKernel` class implementations should use encapsulating
        round brackets to specify their constituents. (v >= 1.0.1)

    :arg name: str. The codename of the desired :code:`_Kernel` instance.

    :returns: object. The desired :code:`_Kernel` instance with default parameters. Returns :code:`None` if given kernel codename was invalid.
    '''

    kernel = None
    if isinstance(name, str):
        m = re.search(r'^(.*?)\((.*)\)$',name)
        if m:
            links = m.group(2).split('-')
            names = []
            bflag = False
            rname = ''
            for jj in range(len(links)):
                rname = links[jj] if not bflag else rname + '-' + links[jj]
                if re.search(r'\(', links[jj]):
                    bflag = True
                if re.search(r'\)', links[jj]):
                    bflag = False
                if not bflag:
                    names.append(rname)
            kklist = []
            for ii in range(len(names)):
                kklist.append(KernelConstructor(names[ii], dtype=dtype))
            if re.search(r'^Sum$', m.group(1)):
                kernel = Sum_Kernel(klist=kklist, dtype=dtype)
            elif re.search(r'^Prod$', m.group(1)):
                kernel = Product_Kernel(klist=kklist, dtype=dtype)
            elif re.search(r'^Sym$', m.group(1)):
                kernel = Symmetric_Kernel(klist=kklist, dtype=dtype)
            elif re.search(r'^NSum$', m.group(1)):
                kernel = ND_Sum_Kernel(klist=kklist, dtype=dtype)
            elif re.search(r'^NProd$', m.group(1)):
                kernel = ND_Product_Kernel(klist=kklist, dtype=dtype)
        else:
            if re.search(r'^C$', name):
                kernel = Constant_Kernel(dtype=dtype)
            elif re.search(r'^n$', name):
                kernel = Noise_Kernel(dtype=dtype)
            elif re.search(r'^L$', name):
                kernel = Linear_Kernel(dtype=dtype)
            elif re.search(r'^P$', name):
                kernel = Poly_Order_Kernel(dtype=dtype)
            elif re.search(r'^SE$', name):
                kernel = SE_Kernel(dtype=dtype)
            elif re.search(r'^RQ$', name):
                kernel = RQ_Kernel(dtype=dtype)
            elif re.search(r'^MH$', name):
                kernel = Matern_HI_Kernel(dtype=dtype)
            elif re.search(r'^NN$', name):
                kernel = NN_Kernel(dtype=dtype)
            elif re.search(r'^Gw', name):
                wname = re.search(r'^Gw(.*)$', name).group(1)
                wfunc = None
                if re.search(r'^C$', wname):
                    wfunc = Constant_WarpingFunction(dtype=dtype)
                elif re.search(r'^L$', wname):
                    wfunc = Linear_WarpingFunction(dtype=dtype)
                elif re.search(r'^IG$', wname):
                    wfunc = IG_WarpingFunction(dtype=dtype)
                kernel = Gibbs_Kernel(wfunc=wfunc, dtype=dtype)
    return kernel


def KernelReconstructor(name, pars=None, dtype=None):
    r'''
    Function to reconstruct any :code:`_Kernel` instance from its codename and parameter list,
    useful for saving only necessary data to represent a :code:`GaussianProcessRregression1D`
    instance.

    .. note::

        All :code:`_OperatorKernel` class implementations should use encapsulating
        round brackets to specify their constituents. (v >= 1.0.1)

    :arg name: str. The codename of the desired :code:`_Kernel` instance.

    :kwarg pars: array. The hyperparameter and constant values to be stored in the :code:`_Kernel` instance, order determined by the specific :code:`_Kernel` class implementation. (optional)

    :returns: object. The desired :code:`_Kernel` instance, with the supplied parameters already set if parameters were valid. Returns :code:`None` if given kernel codename was invalid.
    '''

    dt = dtype if dtype is not None else default_dtype
    kernel = KernelConstructor(name, dtype=dt)
    pvec = None
    if isinstance(pars, array_types):
        pvec = np.array(pars, dtype=dt).flatten()
    #elif isinstance(pars, np.ndarray):
    #    pvec = pars.flatten()
    if isinstance(kernel, _Kernel) and pvec is not None:
        nhyp = kernel.hyperparameters.size
        ncst = kernel.constants.size
        if ncst > 0 and pvec.size >= (nhyp + ncst):
            csts = pvec[nhyp:nhyp+ncst].copy() if pvec.size > (nhyp + ncst) else pvec[nhyp:].copy()
            kernel.constants = csts
        if pvec.size >= nhyp:
            theta = pvec[:nhyp].copy() if pvec.size > nhyp else pvec.copy()
            kernel.hyperparameters = theta
    return kernel


def diagonal(matrix, dtype=None):
    r'''
    Function to compute diagonal of N x [D x ...] x N matrix.

    :returns: array. The desired N x D matrix, which contains the diagonal of the inner matrix in dimension D.
    '''
    dt = dtype if dtype is not None else default_dtype
    mat = None
    diag = None
    if isinstance(matrix, array_types):
        mat = np.array(matrix, dtype=dt)
    if mat is not None and mat.ndim > 1 and mat.shape[0] == mat.shape[-1]:
        if mat.ndim > 2:
            diag = np.zeros((mat.shape[0], mat.shape[1]), dtype=dt)
            for ii in range(mat.shape[0]):
                for jj in range(mat.shape[1]):
                    index = [jj] * (mat.ndim - 2)
                    diag[ii, jj] = mat[(ii, *index, ii)]
        else:
            diag = np.diag(mat)
    return diag


def diagonalize(diagonal, full=True, dtype=None):
    r'''
    Function to compute diagonal matrix from N x D matrix.

    :returns: array. The desired N x [D x ...] x N matrix, which expands the given diagonal across the 2- or D-dimensional inner matrix.
    '''
    dt = dtype if dtype is not None else default_dtype
    diag = None
    mat = None
    if isinstance(diagonal, array_types):
        diag = np.array(diagonal, dtype=dt)
    if diag is not None and diag.ndim > 0 and diag.ndim <= 2:
        if diag.ndim == 2:
            ndim = diag.shape[-1] if full else 2
            dshape = [diag.shape[-1]] * ndim
            mat = np.zeros((diag.shape[0], *dshape, diag.shape[0]), dtype=dt)
            for ii in range(mat.shape[0]):
                for jj in range(mat.shape[1]):
                    element = [jj] * ndim
                    mat[(ii, *element, ii)] = diag[ii, jj]
        elif diag.ndim == 1:
            if full:
                dshape = [diag.shape[0]] * diag.shape[0]
                mat = np.zeros(dshape, dtype=dt)
                for ii in range(mat.shape[0]):
                    element = [ii] * mat.shape[0]
                    mat[(..., *element)] = diag[ii]
            else:
                mat = np.diag(diag)
    return mat



r'''
Some helpful definitions for reproducability and user-friendliness
'''
# Required imports
import numpy as np

np_itypes = (np.int8, np.int16, np.int32, np.int64)
np_utypes = (np.uint8, np.uint16, np.uint32, np.uint64)
np_ftypes = (np.float16, np.float32, np.float64)
number_types = (float, int, np_itypes, np_utypes, np_ftypes)
array_types = (list, tuple, np.ndarray)

default_dtype = np.float64




#!/usr/bin/env python

import pytest
import numpy as np
from mkgp.core.baseclasses import _Kernel
from mkgp.core.utils import KernelConstructor, KernelReconstructor


def check_kernel_evaluation(kernel, x1, x2, der, comparison):
    return np.all(np.isclose(kernel(x1, x2, der=der), comparison))

def check_kernel_transposition(kernel, x1, x2):
    return np.all(np.isclose(kernel(x1, x2, der=1), np.transpose(kernel(x1, x2, der=-1))))

def check_kernel_hyperparameter_derivatives(kernel, x1, x2, comparison):
    hderiv_tests = [False] * len(kernel.hyperparameters)
    if len(comparison) == len(hderiv_tests):
        for ihyp in np.arange(0, len(hderiv_tests)):
            hderiv_tests[ihyp] = np.all(np.isclose(kernel(x1, x2, hder=ihyp), comparison[ihyp]))
    return np.all(hderiv_tests)


@pytest.mark.base_classes
@pytest.mark.usefixtures('empty_warping_function', 'empty_kernel', 'empty_operator_kernel')
class TestBaseClasses():

    def test_empty_warping_function_defaults(self, empty_warping_function):
        assert (empty_warping_function.name,
                empty_warping_function.hyperparameters.size,
                empty_warping_function.constants.size,
                empty_warping_function.bounds,
                empty_warping_function.is_hderiv_implemented()) == ('Warp', 0, 0, None, False)

    def test_eval_empty_warping_function(self, empty_warping_function):
        pytest.raises(NotImplementedError, empty_warping_function, [0.0])

    def test_empty_kernel_defaults(self, empty_kernel):
        assert (empty_kernel.name,
                empty_kernel.hyperparameters.size,
                empty_kernel.constants.size,
                empty_kernel.bounds,
                empty_kernel.is_hderiv_implemented()) == ('None', 0, 0, None, False)

    def test_eval_empty_kernel(self, empty_kernel):
        pytest.raises(NotImplementedError, empty_kernel, [0.0], [0.0])

    def test_empty_operator_kernel_defaults(self, empty_operator_kernel):
        assert (empty_operator_kernel.name,
                empty_operator_kernel.hyperparameters.size,
                empty_operator_kernel.constants.size,
                empty_operator_kernel.bounds,
                empty_operator_kernel.is_hderiv_implemented()) == ('Op()', 0, 0, None, False)

    def test_eval_empty_operator_kernel(self, empty_operator_kernel):
        pytest.raises(NotImplementedError, empty_operator_kernel, [0.0], [0.0])


@pytest.mark.helpers
@pytest.mark.usefixtures('se_kernel', 'gibbs_inverse_gaussian_kernel', 'sum_kernel')
class TestAutoKernelCreation():

    def test_constructor_kernels(self, se_kernel):
        gpobj = KernelConstructor(se_kernel.name)
        assert isinstance(gpobj, _Kernel), f'{gpobj:!r} is not Kernel object!'

    def test_constructor_warping_functions(self, gibbs_constant_kernel):
        gpobj = KernelConstructor(gibbs_constant_kernel.name)
        assert isinstance(gpobj, _Kernel), f'{gpobj:!r} is not Kernel object!'

    def test_constructor_operators(self, sum_kernel):
        gpobj = KernelConstructor(sum_kernel.name)
        assert isinstance(gpobj, _Kernel), f'{gpobj:!r} is not Kernel object!'

    def test_reconstructor_kernels(self, se_kernel):
        kernel_pars = np.hstack((se_kernel.hyperparameters, se_kernel.constants)).flatten()
        gpobj = KernelReconstructor(se_kernel.name, kernel_pars)
        assert gpobj == se_kernel

    def test_reconstructor_warping_functions(self, gibbs_inverse_gaussian_kernel):
        kernel_pars = np.hstack((gibbs_inverse_gaussian_kernel.hyperparameters, gibbs_inverse_gaussian_kernel.constants)).flatten()
        gpobj = KernelReconstructor(gibbs_inverse_gaussian_kernel.name, kernel_pars)
        assert gpobj == gibbs_inverse_gaussian_kernel

    def test_reconstructor_operators(self, sum_kernel):
        kernel_pars = np.hstack((sum_kernel.hyperparameters, sum_kernel.constants)).flatten()
        gpobj = KernelReconstructor(sum_kernel.name, kernel_pars)
        assert gpobj == sum_kernel


@pytest.mark.kernels
@pytest.mark.usefixtures('constant_kernel')
class TestConstantKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.full((x1_vector.size, x2_vector.size), 2.0).T
    ref_dcov = np.zeros((x1_vector.size, x2_vector.size)).T
    ref_ddcov = np.zeros((x1_vector.size, x2_vector.size)).T
    ref_hdcov = [
        np.ones((x1_vector.size, x2_vector.size)).T,
    ]

    def test_eval(self, constant_kernel):
        assert check_kernel_evaluation(constant_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, constant_kernel):
        assert check_kernel_evaluation(constant_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, constant_kernel):
        assert check_kernel_transposition(constant_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, constant_kernel):
        assert check_kernel_evaluation(constant_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self,constant_kernel):
        if constant_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(constant_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, constant_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('noise_kernel')
class TestNoiseKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.diag(np.power(np.ones(x1_vector.shape), 2.0))
    ref_dcov = np.zeros((x1_vector.size, x2_vector.size)).T
    ref_ddcov = np.zeros((x1_vector.size, x2_vector.size)).T
    ref_hdcov = [
        np.diag(2.0 * np.ones(x1_vector.shape)),
    ]

    def test_eval(self, noise_kernel):
        assert check_kernel_evaluation(noise_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self,noise_kernel):
        assert check_kernel_evaluation(noise_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, noise_kernel):
        assert check_kernel_transposition(noise_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, noise_kernel):
        assert check_kernel_evaluation(noise_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, noise_kernel):
        if noise_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(noise_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, noise_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('linear_kernel')
class TestLinearKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[0.0, 0.0], [0.0, 4.0]])
    ref_dcov = np.atleast_2d([[0.0, 4.0], [0.0, 4.0]])
    ref_ddcov = np.atleast_2d([[4.0, 4.0], [4.0, 4.0]])
    ref_hdcov = [
        np.atleast_2d([[0.0, 0.0], [0.0, 4.0]]),
    ]

    def test_eval(self, linear_kernel):
        assert check_kernel_evaluation(linear_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, linear_kernel):
        assert check_kernel_evaluation(linear_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, linear_kernel):
        assert check_kernel_transposition(linear_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, linear_kernel):
        assert check_kernel_evaluation(linear_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, linear_kernel):
        if linear_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(linear_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, linear_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('poly_order_kernel')
class TestPolyOrderKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[1.0, 1.0], [1.0, 5.0]])
    ref_dcov = np.atleast_2d([[0.0, 4.0], [0.0, 4.0]])
    ref_ddcov = np.atleast_2d([[4.0, 4.0], [4.0, 4.0]])
    ref_hdcov = [
        np.atleast_2d([[0.0, 0.0], [0.0, 4.0]]),
        np.atleast_2d([[1.0, 1.0], [1.0, 1.0]]),
    ]

    def test_eval(self, poly_order_kernel):
        assert check_kernel_evaluation(poly_order_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, poly_order_kernel):
        assert check_kernel_evaluation(poly_order_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, poly_order_kernel):
        assert check_kernel_transposition(poly_order_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, poly_order_kernel):
        assert check_kernel_evaluation(poly_order_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, poly_order_kernel):
        if poly_order_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(poly_order_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, poly_order_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('se_kernel')
class TestSquareExponentialKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[1.0, 0.13533528323], [0.13533528323, 1.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.54134113295], [-0.54134113295, 0.0]])
    ref_ddcov = np.atleast_2d([[4.0, -1.62402339884], [-1.62402339884, 4.0]])
    ref_hdcov = [
        np.atleast_2d([[2.0, 0.27067056647], [0.27067056647, 2.0]]),
        np.atleast_2d([[0.0, 1.08268226589], [1.08268226589, 0.0]]),
    ]

    def test_eval(self, se_kernel):
        assert check_kernel_evaluation(se_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, se_kernel):
        assert check_kernel_evaluation(se_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, se_kernel):
        assert check_kernel_transposition(se_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, se_kernel):
        assert check_kernel_evaluation(se_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, se_kernel):
        if se_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(se_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, se_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('rq_kernel')
class TestRationalQuadraticKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[1.0, 0.18593443208], [0.18593443208, 1.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.53124123452], [-0.53124123452, 0.0]])
    ref_ddcov = np.atleast_2d([[4.0, -1.29015728383], [-1.29015728383, 4.0]])
    ref_hdcov = [
        np.atleast_2d([[2.0, 0.37186886416], [0.37186886416, 2.0]]),
        np.atleast_2d([[0.0, 0.74373772833], [0.74373772833, 0.0]]),
        np.atleast_2d([[0.0, -0.00943765078], [-0.00943765078, 0.0]]),
    ]

    def test_eval(self, rq_kernel):
        assert check_kernel_evaluation(rq_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, rq_kernel):
        assert check_kernel_evaluation(rq_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, rq_kernel):
        assert check_kernel_transposition(rq_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, rq_kernel):
        assert check_kernel_evaluation(rq_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, rq_kernel):
        if rq_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(rq_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, rq_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('matern_hi_kernel')
class TestMaternHalfIntegerKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[1.0, 0.13866021914], [0.13866021914, 1.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.41671741677], [-0.41671741677, 0.0]])
    ref_ddcov = np.atleast_2d([[6.66666666667, -1.10633471569], [-1.10633471569, 6.66666666667]])
    ref_hdcov = [
        np.atleast_2d([[2.0, 0.27732043828], [0.27732043828, 2.0]]),
        np.atleast_2d([[0.0, 0.18636169426], [0.18636169426, 0.0]]),
    ]

    def test_eval(self, matern_hi_kernel):
        assert check_kernel_evaluation(matern_hi_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, matern_hi_kernel):
        assert check_kernel_evaluation(matern_hi_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, matern_hi_kernel):
        assert check_kernel_transposition(matern_hi_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, matern_hi_kernel):
        assert check_kernel_evaluation(matern_hi_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, matern_hi_kernel):
        if matern_hi_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(matern_hi_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, matern_hi_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('gibbs_constant_kernel')
class TestGibbsKernelWithConstant():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[1.0, 0.60653065971], [0.60653065971, 1.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.60653065971], [-0.60653065971, 0.0]])
    ref_ddcov = np.atleast_2d([[1.0, 0.0], [0.0, 1.0]])
    ref_hdcov = [
        np.atleast_2d([[2.0, 1.21306131943], [1.21306131943, 2.0]]),
        np.atleast_2d([[0.0, 0.60653065971], [0.60653065971, 0.0]]),
    ]

    def test_eval(self, gibbs_constant_kernel):
        assert check_kernel_evaluation(gibbs_constant_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, gibbs_constant_kernel):
        assert check_kernel_evaluation(gibbs_constant_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, gibbs_constant_kernel):
        assert check_kernel_transposition(gibbs_constant_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, gibbs_constant_kernel):
        assert check_kernel_evaluation(gibbs_constant_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, gibbs_constant_kernel):
        if gibbs_constant_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(gibbs_constant_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, gibbs_constant_kernel, **kwargs)


@pytest.mark.kernels
@pytest.mark.usefixtures('gibbs_inverse_gaussian_kernel')
class TestGibbsKernelWithInverseGaussian():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[1.0, 0.10737182452], [0.10737182452, 1.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.47848791644], [0.58059409716, 0.0]])
    ref_ddcov = np.atleast_2d([[4.0, 1.00715576945], [1.00715576945, 16.8232521516]])
    ref_hdcov = [
        np.atleast_2d([[2.0, 0.21474364904], [0.21474364904, 2.0]]),
        np.atleast_2d([[-2.0, 1.88160836531], [1.88160836531, -2.78744556822]]),
        np.atleast_2d([[0.0, -0.12820614183], [-0.12820614183, 0.37723973548]]),
        np.atleast_2d([[0.0, -4.10259653856], [-4.10259653856, 12.0716715354]]),
    ]

    def test_eval(self, gibbs_inverse_gaussian_kernel):
        assert check_kernel_evaluation(gibbs_inverse_gaussian_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, gibbs_inverse_gaussian_kernel):
        assert check_kernel_evaluation(gibbs_inverse_gaussian_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, gibbs_inverse_gaussian_kernel):
        assert check_kernel_transposition(gibbs_inverse_gaussian_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, gibbs_inverse_gaussian_kernel):
        assert check_kernel_evaluation(gibbs_inverse_gaussian_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, gibbs_inverse_gaussian_kernel):
        if gibbs_inverse_gaussian_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(gibbs_inverse_gaussian_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, gibbs_inverse_gaussian_kernel, **kwargs)


@pytest.mark.operator_kernels
@pytest.mark.usefixtures('sum_kernel')
class TestSumOperationKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[2.0, 0.13533528323], [0.13533528323, 2.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.54134113295], [-0.54134113295, 0.0]])
    ref_ddcov = np.atleast_2d([[4.0, -1.62402339884], [-1.62402339884, 4.0]])
    ref_hdcov = [
        np.atleast_2d([[3.0, 0.27067056647], [0.27067056647, 3.0]]),
        np.atleast_2d([[1.0, 1.08268226589], [1.08268226589, 1.0]]),
        np.atleast_2d([[3.0, 0.13533528324], [0.13533528324, 3.0]]),
    ]

    def test_eval(self, sum_kernel):
        assert check_kernel_evaluation(sum_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, sum_kernel):
        assert check_kernel_evaluation(sum_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, sum_kernel):
        assert check_kernel_transposition(sum_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, sum_kernel):
        assert check_kernel_evaluation(sum_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, sum_kernel):
        if sum_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(sum_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, sum_kernel, **kwargs)


@pytest.mark.operator_kernels
@pytest.mark.usefixtures('product_kernel')
class TestProductOperationKernel():

    x1_vector = np.array([0.0, 1.0])
    x2_vector = np.array([0.0, 1.0])
    ref_cov = np.atleast_2d([[0.0, 0.0], [0.0, 16.0]])
    ref_dcov = np.atleast_2d([[0.0, 0.0], [0.0, 32.0]])
    ref_ddcov = np.atleast_2d([[0.0, 0.0], [0.0, 64.0]])
    ref_hdcov = [
        np.atleast_2d([[0.0, 0.0], [0.0, 16.0]]),
        np.atleast_2d([[0.0, 0.0], [0.0, 16.0]]),
    ]

    def test_eval(self, product_kernel):
        assert check_kernel_evaluation(product_kernel, self.x1_vector, self.x2_vector, 0, self.ref_cov)

    def test_eval_first_derivative(self, product_kernel):
        assert check_kernel_evaluation(product_kernel, self.x1_vector, self.x2_vector, 1, self.ref_dcov)

    def test_eval_first_derivative_transpose(self, product_kernel):
        assert check_kernel_transposition(product_kernel, self.x1_vector, self.x2_vector)

    def test_eval_second_derivative(self, product_kernel):
        assert check_kernel_evaluation(product_kernel, self.x1_vector, self.x2_vector, 2, self.ref_ddcov)

    def test_eval_hyperparameter_derivatives(self, product_kernel):
        if product_kernel.is_hderiv_implemented():
            assert check_kernel_hyperparameter_derivatives(product_kernel, self.x1_vector, self.x2_vector, self.ref_hdcov)
        else:
            kwargs = {'x1': self.x1_vector, 'x2': self.x2_vector, 'hder': 0}
            pytest.raises(NotImplementedError, product_kernel, **kwargs)
